/*
 * Renderer 8. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

package renderer.pipeline;

import renderer.scene.*;
import renderer.scene.util.CheckModel;
import renderer.framebuffer.*;
import static renderer.pipeline.PipelineLogger.*;

import java.awt.Color;

/**
   This renderer takes as its input a {@link Scene} data structure
   and a {@link FrameBuffer.Viewport} within a {@link FrameBuffer}
   data structure. This renderer mutates the {@link FrameBuffer.Viewport}
   so that it is filled in with the rendered image of the geometric
   scene represented by the {@link Scene} object.
<p>
   This implements our seventh rendering pipeline. In the previous
   renderer the first pipeline stage was Model2View. This renderer
   splits that stage into two stages, {@link Model2World} and
   {@link World2View}. The reason for splitting the first stage is
   that we have a new coordinate system, world coordinates, that is
   logically between model coordinates and view coordinates. Recall
   that "model coordinates" are relative to a model's private origin
   and "view coordinates" are relative to the camera's arbitrary view
   volume. World coordinates are relative to a fixed origin shared by
   every object in the scene, including the camera. In this renderer
   the camera can be placed anywhere in the world coordinate system.
   In other words, the world coordinate system facilitates moving the
   {@link Camera} around in a scene. There are now seven pipeline stages.
*/
public final class Pipeline
{
   // Mostly for compatibility with renderers 1 through 3.
   public static Color DEFAULT_COLOR = Color.white;

   /**
      Mutate the {@link FrameBuffer}'s default {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer fb)
   {
      render(scene, fb.vp); // render into the default viewport
   }


   /**
      Mutate the {@link FrameBuffer}'s given {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param vp     {@link FrameBuffer.Viewport} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer.Viewport vp)
   {
      PipelineLogger.debugScene = scene.debug;

      logMessage("\n== Begin Rendering of Scene: " + scene.name + " ==");

      logMessage("-- Current Camera:\n" + scene.camera);

      // For every Position in the Scene, render the Position's Model.
      for (final Position position : scene.positionList)
      {
         PipelineLogger.debugPosition = position.debug;

         if ( position.visible )
         {
            logMessage("==== Render position: " + position.name + " ====");

            logMessage("---- Translation vector = " + position.getTranslation());

            if ( position.getModel().visible )
            {
               logMessage("====== Render model: "
                                  + position.getModel().name + " ======");

               CheckModel.check(position.getModel());

               // Mostly for compatibility with renderers 1 through 3.
               if (  position.getModel().colorList.isEmpty()
                 && !position.getModel().vertexList.isEmpty())
               {
                  for (int i = 0; i < position.getModel().vertexList.size(); ++i)
                  {
                     position.getModel().addColor( DEFAULT_COLOR );
                  }
                  System.err.println("***WARNING: Added default color to model: "
                                    + position.getModel().name + ".");
               }

               logVertexList("0. Model        ", position.getModel());

               // 1. Apply the Position's model-to-world coordinate transformation.
               final Model model1 = Model2World.model2world(position);

               logVertexList("1. World        ", model1);

               // 2. Apply the Camera's world-to-view coordinate transformation.
               final Model model2 = World2View.world2view(model1,
                                                          scene.camera);

               logVertexList("2. View         ", model2);

               // 3. Apply the Camera's normalizing view-to-camera coordinate transformation.
               final Model model3 = View2Camera.view2camera(model2,
                                                            scene.camera);

               logVertexList("3. Camera       ", model3);
               logColorList("3. Camera       ", model3);
               logPrimitiveList("3. Camera       ", model3);

               // 4. Clip primitives to the camera's near plane.
               final Model model4 = NearClip.clip(model3,
                                                  scene.camera);

               logVertexList("4. Near_Clipped", model4);
               logColorList("4. Near_Clipped", model4);
               logPrimitiveList("4. Near_Clipped", model4);

               // 5. Apply the Camera's projection transformation.
               final Model model5 = Projection.project(model4,
                                                       scene.camera);

               logVertexList("5. Projected   ", model5);

               // 6. Clip primitives to the camera's view rectangle.
               final Model model6 = Clip.clip(model5);

               logVertexList("6. Clipped     ", model6);
               logColorList("6. Clipped     ", model6);
               logPrimitiveList("6. Clipped     ", model6);

               // 7. Rasterize every visible primitive into pixels.
               Rasterize.rasterize(model6, vp);

               logMessage("====== End model: "
                                  + position.getModel().name + " ======");
            }
            else
            {
               logMessage("====== Hidden model: "
                                  + position.getModel().name + " ======");
            }

            logMessage("==== End position: " + position.name + " ====");
         }
         else
         {
            logMessage("==== Hidden position: " + position.name + " ====");
         }
      }
      logMessage("== End Rendering of Scene ==");
   }



   // Private default constructor to enforce noninstantiable class.
   // See Item 4 in "Effective Java", 3rd Ed, Joshua Bloch.
   private Pipeline() {
      throw new AssertionError();
   }
}
